@@ -120,6 +120,10 @@ enum {
 	PROP_STATUS_CODE,
 	PROP_REASON_PHRASE,
 	PROP_FIRST_PARTY,
+	PROP_REQUEST_BODY,
+	PROP_REQUEST_HEADERS,
+	PROP_RESPONSE_BODY,
+	PROP_RESPONSE_HEADERS,
 
 	LAST_PROP
 };
@@ -416,7 +420,7 @@ soup_message_class_init (SoupMessageClass *message_class)
 	 * SoupMessage::content-sniffed:
 	 * @msg: the message
 	 * @type: the content type that we got from sniffing
-	 * @params: a #GHashTable with the parameters
+	 * @params: (element-type utf8 utf8): a #GHashTable with the parameters
 	 *
 	 * This signal is emitted after %got-headers, and before the
 	 * first %got-chunk. If content sniffing is disabled, or no
@@ -579,7 +583,6 @@ soup_message_class_init (SoupMessageClass *message_class)
 				     "The HTTP response reason phrase",
 				     NULL,
 				     G_PARAM_READWRITE));
-
 	/**
 	 * SOUP_MESSAGE_FIRST_PARTY:
 	 *
@@ -594,6 +597,58 @@ soup_message_class_init (SoupMessageClass *message_class)
 				    "The URI loaded in the application when the message was requested.",
 				    SOUP_TYPE_URI,
 				    G_PARAM_READWRITE));
+	/**
+	 * SOUP_MESSAGE_REQUEST_BODY:
+	 *
+	 * Alias for the #SoupMessage:request-body property. (The
+	 * message's HTTP request body.)
+	 **/
+	g_object_class_install_property (
+		object_class, PROP_REQUEST_BODY,
+		g_param_spec_boxed (SOUP_MESSAGE_REQUEST_BODY,
+				    "Request Body",
+				    "The HTTP request content",
+				    SOUP_TYPE_MESSAGE_BODY,
+				    G_PARAM_READABLE));
+	/**
+	 * SOUP_MESSAGE_REQUEST_HEADERS:
+	 *
+	 * Alias for the #SoupMessage:request-headers property. (The
+	 * message's HTTP request headers.)
+	 **/
+	g_object_class_install_property (
+		object_class, PROP_REQUEST_HEADERS,
+		g_param_spec_boxed (SOUP_MESSAGE_REQUEST_HEADERS,
+				    "Request Headers",
+				    "The HTTP request headers",
+				    SOUP_TYPE_MESSAGE_HEADERS,
+				    G_PARAM_READABLE));
+	/**
+	 * SOUP_MESSAGE_RESPONSE_BODY:
+	 *
+	 * Alias for the #SoupMessage:response-body property. (The
+	 * message's HTTP response body.)
+	 **/
+	g_object_class_install_property (
+		object_class, PROP_RESPONSE_BODY,
+		g_param_spec_boxed (SOUP_MESSAGE_RESPONSE_BODY,
+				    "Response Body",
+				    "The HTTP response content",
+				    SOUP_TYPE_MESSAGE_BODY,
+				    G_PARAM_READABLE));
+	/**
+	 * SOUP_MESSAGE_RESPONSE_HEADERS:
+	 *
+	 * Alias for the #SoupMessage:response-headers property. (The
+	 * message's HTTP response headers.)
+	 **/
+	g_object_class_install_property (
+		object_class, PROP_RESPONSE_HEADERS,
+		g_param_spec_boxed (SOUP_MESSAGE_RESPONSE_HEADERS,
+				    "Response Headers",
+				     "The HTTP response headers",
+				    SOUP_TYPE_MESSAGE_HEADERS,
+				    G_PARAM_READABLE));
 }
 
 static void
@@ -671,6 +726,18 @@ get_property (GObject *object, guint prop_id,
 	case PROP_FIRST_PARTY:
 		g_value_set_boxed (value, priv->first_party);
 		break;
+	case PROP_REQUEST_BODY:
+		g_value_set_boxed (value, msg->request_body);
+		break;
+	case PROP_REQUEST_HEADERS:
+		g_value_set_boxed (value, msg->request_headers);
+		break;
+	case PROP_RESPONSE_BODY:
+		g_value_set_boxed (value, msg->response_body);
+		break;
+	case PROP_RESPONSE_HEADERS:
+		g_value_set_boxed (value, msg->response_headers);
+		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
@@ -764,9 +831,10 @@ soup_message_set_request (SoupMessage    *msg,
 /**
  * soup_message_set_response:
  * @msg: the message
- * @content_type: MIME Content-Type of the body
+ * @content_type: (allow-none): MIME Content-Type of the body
  * @resp_use: a #SoupMemoryUse describing how to handle @resp_body
- * @resp_body: a data buffer containing the body of the message response.
+ * @resp_body: (array length=resp_length) (element-type uint8): a data buffer
+ * containing the body of the message response.
  * @resp_length: the byte length of @resp_body.
  * 
  * Convenience function to set the response body of a #SoupMessage. If
@@ -1258,6 +1326,7 @@ soup_message_cleanup_response (SoupMessage *req)
 		priv->decoders = g_slist_delete_link (priv->decoders, priv->decoders);
 	}
 	priv->msg_flags &= ~SOUP_MESSAGE_CONTENT_DECODED;
+	priv->msg_flags &= ~SOUP_MESSAGE_CERTIFICATE_TRUSTED;
 
 	req->status_code = SOUP_STATUS_NONE;
 	if (req->reason_phrase) {
@@ -1279,6 +1348,13 @@ soup_message_cleanup_response (SoupMessage *req)
  * @SOUP_MESSAGE_OVERWRITE_CHUNKS: Deprecated: equivalent to calling
  * soup_message_body_set_accumulate() on the incoming message body
  * (ie, %response_body for a client-side request), passing %FALSE.
+ * @SOUP_MESSAGE_CONTENT_DECODED: Set by #SoupContentDecoder to
+ * indicate that it has removed the Content-Encoding on a message (and
+ * so headers such as Content-Length may no longer accurately describe
+ * the body).
+ * @SOUP_MESSAGE_CERTIFICATE_TRUSTED: if %TRUE after an https response
+ * has been received, indicates that the server's SSL certificate is
+ * trusted according to the session's CA.
  *
  * Various flags that can be set on a #SoupMessage to alter its
  * behavior.
